/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2008 Digium
 *
 * Adapted from cdr_adaptive_odbc:
 * Tilghman Lesher <tlesher AT digium DOT com>
 * by Steve Murphy
 *
 * Refactored most of this module to support delayed connectivity.
 * Dennis Buteyn <dennis.buteyn@xorcom.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief ODBC CEL backend
 *
 * \author Tilghman Lesher \verbatim <tlesher AT digium DOT com> \endverbatim
 * \ingroup cel_drivers
 */

/*** MODULEINFO
	<depend>res_odbc</depend>
	<depend>generic_odbc</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision$")

#include <sys/types.h>
#include <time.h>
#include <math.h>

#include <sql.h>
#include <sqlext.h>
#include <sqltypes.h>

#include "asterisk/config.h"
#include "asterisk/channel.h"
#include "asterisk/lock.h"
#include "asterisk/linkedlists.h"
#include "asterisk/res_odbc.h"
#include "asterisk/cel.h"
#include "asterisk/module.h"
#include "asterisk/sem.h"

#define	CONFIG	"cel_odbc.conf"

#define ODBC_BACKEND_NAME "ODBC CEL backend"

/*! \brief show_user_def is off by default */
#define CEL_SHOW_USERDEF_DEFAULT	0

/*! TRUE if we should set the eventtype field to USER_DEFINED on user events. */
static unsigned char cel_show_user_def;

struct columns {
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(name);
	);
	const char * staticval; /* Reference to staticval->value, no need to free */
	int rva; /* Relative offset into ast_cel_event_record */
	AST_LIST_ENTRY(columns) list;
};

struct aliases {
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(source);
		AST_STRING_FIELD(target);
	);
	int rva; /* Relative offset into ast_cel_event_record */
	AST_LIST_ENTRY(aliases) list;
};

struct staticvals {
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(colname);
		AST_STRING_FIELD(value);
	);
	AST_LIST_ENTRY(staticvals) list;
};

struct filters {
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(celname);
		AST_STRING_FIELD(value);
	);
	int rva; /* Relative offset into ast_cel_event_record */
	AST_LIST_ENTRY(filters) list;
};

struct tables {
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(name);
		AST_STRING_FIELD(category);
		AST_STRING_FIELD(connection);
		AST_STRING_FIELD(query);
		AST_STRING_FIELD(quote_char);
	);
	unsigned int usegmtime:1;
	unsigned int allowleapsec:1;
	AST_LIST_HEAD_NOLOCK(odbc_columns, columns) columns;
	AST_LIST_HEAD_NOLOCK(odbc_aliases, aliases) aliases;
	AST_LIST_HEAD_NOLOCK(odbc_staticvals, staticvals) staticvals;
	AST_LIST_HEAD_NOLOCK(odbc_filters, filters) filters;
	AST_RWLIST_ENTRY(tables) list;
};

/* Used when calling generic_prepare */
struct prepare_params {
	struct tables *table;
	struct ast_cel_event_record record;
	SQL_TIMESTAMP_STRUCT event_time;
};

static AST_RWLIST_HEAD_STATIC(odbc_tables, tables);
struct ast_sem manager_sem; /* Raised when manager told to stop */
static pthread_t manager_thread = AST_PTHREADT_NULL;

static struct columns * create_column(void)
{
	struct columns *column = ast_calloc(1, sizeof(*column));

	if (column) {
		if (ast_string_field_init(column, 256) == 0) {
			column->rva = -1;
			return column;
		} else {
			ast_free(column);
		}
	}

	return NULL;
}

static void destroy_column(struct columns *column)
{
	ast_string_field_free_memory(column);
	ast_free(column);
}

static void destroy_columns(struct tables *table)
{
	struct columns *column;

	while ((column = AST_LIST_REMOVE_HEAD(&(table->columns), list))) {
		destroy_column(column);
	}
}

static struct aliases * create_alias(void)
{
	struct aliases *alias = ast_calloc(1, sizeof(*alias));

	if (alias) {
		if (ast_string_field_init(alias, 256) == 0) {
			alias->rva = -1;
			return alias;
		} else {
			ast_free(alias);
		}
	}

	return NULL;
}

static void destroy_alias(struct aliases *alias)
{
	ast_string_field_free_memory(alias);
	ast_free(alias);
}

static void destroy_aliases(struct tables *table)
{
	struct aliases *alias;

	while ((alias = AST_RWLIST_REMOVE_HEAD(&(table->aliases), list))) {
		destroy_alias(alias);
	}
}

static int append_alias(
	struct tables *table,
	const char *source,
	const char *target,
	int rva
) {
	struct aliases *alias = create_alias();

	if (alias) {
		ast_string_field_set(alias, source, source);
		ast_string_field_set(alias, target, target);
		alias->rva = rva;
		AST_LIST_INSERT_TAIL(&(table->aliases), alias, list);
		return 1;
	} else {
		return 0;
	}
}

static struct staticvals * create_staticval(void)
{
	struct staticvals *staticval = ast_calloc(1, sizeof(*staticval));

	if (staticval) {
		if (ast_string_field_init(staticval, 256) == 0) {
			return staticval;
		} else {
			ast_free(staticval);
		}
	}

	return NULL;
}

static void destroy_staticval(struct staticvals *staticval)
{
	ast_string_field_free_memory(staticval);
	ast_free(staticval);
}

static void destroy_staticvals(struct tables *table)
{
	struct staticvals *staticval;

	while ((staticval = AST_RWLIST_REMOVE_HEAD(&(table->staticvals), list))) {
		destroy_staticval(staticval);
	}
}

static int append_staticval(
	struct tables *table,
	const char *colname,
	const char *value
) {
	struct staticvals *staticval = create_staticval();

	if (staticval) {
		ast_string_field_set(staticval, colname, colname);
		ast_string_field_set(staticval, value, value);
		AST_LIST_INSERT_TAIL(&(table->staticvals), staticval, list);
		return 1;
	} else {
		return 0;
	}
}

static struct filters * create_filter(void)
{
	struct filters *filter = ast_calloc(1, sizeof(*filter));

	if (filter) {
		if (ast_string_field_init(filter, 256) == 0) {
			filter->rva = -1;
			return filter;
		} else {
			ast_free(filter);
		}
	}

	return NULL;
}

static void destroy_filter(struct filters *filter)
{
	ast_string_field_free_memory(filter);
	ast_free(filter);
}

static void destroy_filters(struct tables *table)
{
	struct filters *filter;

	while ((filter = AST_RWLIST_REMOVE_HEAD(&(table->filters), list))) {
		destroy_filter(filter);
	}
}

static int append_filter(
	struct tables *table,
	const char *celname,
	const char *value,
	int rva
) {
	struct filters *filter = create_filter();

	if (filter) {
		ast_string_field_set(filter, celname, celname);
		ast_string_field_set(filter, value, value);
		filter->rva = rva;
		AST_LIST_INSERT_TAIL(&(table->filters), filter, list);
		return 1;
	} else {
		return 0;
	}
}

static struct tables * create_table(void)
{
	struct tables *table = ast_calloc(1, sizeof(*table));

	if (table) {
		if (ast_string_field_init(table, 256) == 0) {
			table->usegmtime = 0;
			table->allowleapsec = 1;
			ast_string_field_set(table, name, "cel");
			ast_string_field_set(table, quote_char, "`");
			return table;
		} else {
			ast_free(table);
		}
	}

	return NULL;
}

static void destroy_table(struct tables *table)
{
	destroy_columns(table);
	destroy_aliases(table);
	destroy_staticvals(table);
	destroy_filters(table);
	ast_string_field_free_memory(table);
	ast_free(table);
}

static void destroy_tables(void)
{
	struct tables *table;

	while ((table = AST_RWLIST_REMOVE_HEAD(&odbc_tables, list))) {
		destroy_table(table);
	}
}

static void destroy_tables_safely(void)
{
	while (AST_RWLIST_WRLOCK(&odbc_tables)) {
		ast_log(LOG_WARNING, "Unable to lock table list for modification. Retrying.\n");
	}
	destroy_tables();
	AST_RWLIST_UNLOCK(&odbc_tables);
}

static int wait_for_signal(void)
{
	struct timeval timeout_val = ast_tvadd(ast_tvnow(), ast_tv(10, 0));
	struct timespec timeout_spec = {
		.tv_sec = timeout_val.tv_sec,
		.tv_nsec = timeout_val.tv_usec * 1000,
	};

	if (ast_sem_timedwait(&manager_sem, &timeout_spec) != 0) {
		return errno;
	} else {
		return 0;
	}
}

static int cel_record_rva(const char * name)
{
	if (ast_strings_equal(name, "eventtime")) {
		return offsetof(struct ast_cel_event_record, event_time);
	} else if (ast_strings_equal(name, "userdeftype")) {
		return offsetof(struct ast_cel_event_record, user_defined_name);
	} else if (ast_strings_equal(name, "cid_name")) {
		return offsetof(struct ast_cel_event_record, caller_id_name);
	} else if (ast_strings_equal(name, "cid_num")) {
		return offsetof(struct ast_cel_event_record, caller_id_num);
	} else if (ast_strings_equal(name, "cid_ani")) {
		return offsetof(struct ast_cel_event_record, caller_id_ani);
	} else if (ast_strings_equal(name, "cid_rdnis")) {
		return offsetof(struct ast_cel_event_record, caller_id_rdnis);
	} else if (ast_strings_equal(name, "cid_dnid")) {
		return offsetof(struct ast_cel_event_record, caller_id_dnid);
	} else if (ast_strings_equal(name, "exten")) {
		return offsetof(struct ast_cel_event_record, extension);
	} else if (ast_strings_equal(name, "context")) {
		return offsetof(struct ast_cel_event_record, context);
	} else if (ast_strings_equal(name, "channame")) {
		return offsetof(struct ast_cel_event_record, channel_name);
	} else if (ast_strings_equal(name, "appname")) {
		return offsetof(struct ast_cel_event_record, application_name);
	} else if (ast_strings_equal(name, "appdata")) {
		return offsetof(struct ast_cel_event_record, application_data);
	} else if (ast_strings_equal(name, "accountcode")) {
		return offsetof(struct ast_cel_event_record, account_code);
	} else if (ast_strings_equal(name, "peeraccount")) {
		return offsetof(struct ast_cel_event_record, peer_account);
	} else if (ast_strings_equal(name, "uniqueid")) {
		return offsetof(struct ast_cel_event_record, unique_id);
	} else if (ast_strings_equal(name, "linkedid")) {
		return offsetof(struct ast_cel_event_record, linked_id);
	} else if (ast_strings_equal(name, "userfield")) {
		return offsetof(struct ast_cel_event_record, user_field);
	} else if (ast_strings_equal(name, "peer")) {
		return offsetof(struct ast_cel_event_record, peer);
	} else if (ast_strings_equal(name, "amaflags")) {
		return offsetof(struct ast_cel_event_record, amaflag);
	} else if (ast_strings_equal(name, "extra")) {
		return offsetof(struct ast_cel_event_record, extra);
	} else if (ast_strings_equal(name, "eventtype")) {
		return offsetof(struct ast_cel_event_record, event_type);
	} else {
		return -1;
	}
}

static void construct_query(struct tables *table)
{
	struct columns *column;
	struct ast_str *query = ast_str_create(2000);
	int count = 0, i;

	ast_str_append(&query, 0, "INSERT INTO %s%s%s (",
		table->quote_char, table->name, table->quote_char);

	AST_LIST_TRAVERSE(&(table->columns), column, list) {
		if (column->rva >= 0 || column->staticval) {
			if (count > 0) {
				ast_str_append(&query, 0, ", %s%s%s",
					table->quote_char, column->name, table->quote_char);
			} else {
				ast_str_append(&query, 0, "%s%s%s",
					table->quote_char, column->name, table->quote_char);
			}
			count++;
		}
	}

	if (count == 0) {
		/* Cannot construct query with zero values to insert. */
		ast_free(query);
		return;
	}

	ast_str_append(&query, 0, "%s", ") VALUES (");

	for (i = 0; i < count; ++i) {
		if (i > 0) {
			ast_str_append(&query, 0, "%s", ", ?");
		} else {
			ast_str_append(&query, 0, "%s", "?");
		}
	}

	ast_str_append(&query, 0, "%s", ")");
	ast_string_field_set(table, query, ast_str_buffer(query));
	ast_free(query);
	ast_verb(3, "Constructed query for %s@%s: %s\n", table->name, table->connection, table->query);
}

static int query_table_info(struct tables *table)
{
	struct odbc_obj *obj;
	SQLHSTMT stmt = NULL;
	struct columns *column;
	char quote_char[10];
	char columnname[80];
	int res;

	obj = ast_odbc_request_obj(table->connection, 1);
	if (!obj) {
		ast_log(LOG_WARNING, "Connection '%s' unavailable at this time.\n", table->connection);
		return 0;
	}

	res = SQLAllocHandle(SQL_HANDLE_STMT, obj->con, &stmt);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		ast_log(LOG_WARNING, "SQL Alloc Handle failed on connection '%s'!\n", table->connection);
		ast_odbc_release_obj(obj);
		return 0;
	}

	res = SQLGetInfo(obj->con, SQL_IDENTIFIER_QUOTE_CHAR, &quote_char, sizeof(quote_char), NULL);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		ast_log(LOG_ERROR, "Unable to determine identifier delimiter on connection '%s'.\n", table->connection);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		ast_odbc_release_obj(obj);
		return 0;
	} else {
		ast_string_field_set(table, quote_char, quote_char);
	}

	res = SQLColumns(stmt, NULL, 0, NULL, 0, (SQLCHAR*) table->name, SQL_NTS, NULL, 0);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		ast_log(LOG_ERROR, "Unable to query database columns on connection '%s'.\n", table->connection);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		ast_odbc_release_obj(obj);
		return 0;
	}

	while (1) {
		res = SQLFetch(stmt);

		if (res == SQL_NO_DATA) {
			/* No more columns, all done */
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			ast_odbc_release_obj(obj);
			return 1;
		} else if (res == SQL_SUCCESS || res == SQL_SUCCESS_WITH_INFO) {
			/* Got a column, save it */
			column = create_column();
			if (column) {
				SQLGetData(stmt, 4, SQL_C_CHAR, columnname, sizeof(columnname), NULL);
				ast_string_field_set(column, name, columnname);
				column->rva = cel_record_rva(columnname);
				AST_LIST_INSERT_TAIL(&(table->columns), column, list);
			} else {
				/* Out of memory, bail out */
				destroy_columns(table);
				SQLFreeHandle(SQL_HANDLE_STMT, stmt);
				ast_odbc_release_obj(obj);
				return 0;
			}
		} else {
			/* Something bad happened, bail out */
			destroy_columns(table);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			ast_odbc_release_obj(obj);
			return 0;
		}
	}
}

static void load_general_config(struct ast_config *cfg)
{
	struct ast_variable *var;

	cel_show_user_def = CEL_SHOW_USERDEF_DEFAULT;

	for (var = ast_variable_browse(cfg, "general"); var; var = var->next) {
		if (strcasecmp(var->name, "show_user_defined") == 0) {
			cel_show_user_def = ast_true(var->value) ? 1 : 0;
		}
	}
}

static char * strip_quotes(char *input)
{
	/* Only to be used with stack-allocated memory! */
	if (input[0] == '"' && input[strlen(input) - 1] == '"') {
		input[strlen(input) - 1] = '\0';
		return input + 1;
	}
	return input;
}

static int load_table_config(
	struct ast_config *cfg,
	const char *category
) {
	struct tables *table;
	struct ast_variable *var;
	int rva;

	table = create_table();
	if (table == NULL) {
		return 0;
	}

	ast_string_field_set(table, category, category);

	for (var = ast_variable_browse(cfg, category); var; var = var->next) {
		if (strcasecmp(var->name, "connection") == 0) {
			ast_string_field_set(table, connection, var->value);
		} else if (strcasecmp(var->name, "table") == 0) {
			ast_string_field_set(table, name, var->value);
		} else if (strcasecmp(var->name, "usegmtime") == 0) {
			table->usegmtime = ast_true(var->value);
		} else if (strcasecmp(var->name, "allowleapsecond") == 0) {
			table->allowleapsec = ast_true(var->value);
		} else if (strncmp(var->name, "alias", 5) == 0) {
			char *source = ast_strip(ast_strdupa(var->name + 5));
			rva = cel_record_rva(source);
			if (rva >= 0) {
				if (!append_alias(table, source, var->value, rva)) {
					destroy_table(table);
					return 0;
				}
			}
		} else if (strncmp(var->name, "static", 6) == 0) {
			char *value = strip_quotes(ast_strip(ast_strdupa(var->name + 6)));
			if (!append_staticval(table, var->value, value)) {
				destroy_table(table);
				return 0;
			}
		} else if (strncmp(var->name, "filter", 6) == 0) {
			char *celname = ast_strip(ast_strdupa(var->name + 6));
			rva = cel_record_rva(celname);
			if (rva >= 0) {
				if (append_filter(table, celname, var->value, rva)) {
					ast_verb(3, "Filtering CEL events matching %s = \"%s\" in %s@%s\n", celname, var->value, table->name, table->connection);
				} else {
					destroy_table(table);
					return 0;
				}
			}
		}
	}

	if (ast_strlen_zero(table->connection)) {
		ast_log(LOG_WARNING, "No connection parameter found in '%s'.  Skipping.\n", category);
		destroy_table(table);
	} else {
		ast_verb(3, "Found CEL table %s@%s.\n", table->name, table->connection);
		AST_RWLIST_INSERT_TAIL(&odbc_tables, table, list);
	}

	return 1;
}

static void resolve_aliases(struct tables *table)
{
	struct columns *column;
	struct aliases *alias;

	/* Somewhat inefficient but not like we're doing this every insertion. */
	AST_LIST_TRAVERSE(&(table->aliases), alias, list) {
		AST_LIST_TRAVERSE(&(table->columns), column, list) {
			if (ast_strings_equal(column->name, alias->target)) {
				column->rva = alias->rva;
				ast_verb(3, "Aliasing %s to %s in %s@%s\n", alias->source, alias->target, table->name, table->connection);
				break;
			}
		}
	}
}

static void resolve_staticvals(struct tables *table)
{
	struct columns *column;
	struct staticvals *staticval;

	/* Somewhat inefficient but not like we're doing this every insertion. */
	AST_LIST_TRAVERSE(&(table->staticvals), staticval, list) {
		AST_LIST_TRAVERSE(&(table->columns), column, list) {
			if (ast_strings_equal(column->name, staticval->colname)) {
				column->staticval = staticval->value;
				ast_verb(3, "Forcing %s to \"%s\" in %s@%s\n", staticval->colname, staticval->value, table->name, table->connection);
				break;
			}
		}
	}
}

static int construct_queries(void)
{
	struct tables *table;
	int failures = 0;

	AST_LIST_TRAVERSE(&odbc_tables, table, list) {
		if (!ast_strlen_zero(table->query)) {
			continue;
		} else if (query_table_info(table)) {
			resolve_aliases(table);
			resolve_staticvals(table);
			construct_query(table);
		} else {
			failures += 1;
		}
	}

	return failures;
}

static int load_config(void)
{
	struct ast_config *cfg;
	const char *catg;

	struct ast_flags config_flags = { 0 }; /* Part of our config comes from the database */

	cfg = ast_config_load(CONFIG, config_flags);
	if (!cfg || cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_WARNING, "Unable to load " CONFIG ".  No ODBC CEL records!\n");
		return -1;
	}

	for (catg = ast_category_browse(cfg, NULL); catg; catg = ast_category_browse(cfg, catg)) {
		if (strcasecmp(catg, "general") == 0) {
			load_general_config(cfg);
		} else if (!load_table_config(cfg, catg)) {
			destroy_tables();
			ast_config_destroy(cfg);
			return -1;
		}
	}

	ast_config_destroy(cfg);

	return construct_queries();
}

static void *mb_manager_thread(void *data)
{
	ast_verb(3, "CEL ODBC Manager started.\n");

	while (1) {
		if (wait_for_signal() == ETIMEDOUT) {
			while (AST_RWLIST_WRLOCK(&odbc_tables)) {
				ast_log(LOG_WARNING, "Unable to lock table list for modification. Retrying.\n");
			}
			construct_queries();
			AST_RWLIST_UNLOCK(&odbc_tables);
		} else {
			ast_verb(3, "Terminating CEL ODBC Manager.\n");
			return NULL;
		}
	}

	return NULL;
}

static int start_table_manager(void)
{
	int errcode = 0;

	ast_sem_init(&manager_sem, 0, 0);

	if ((errcode = ast_pthread_create(&manager_thread, NULL, mb_manager_thread, NULL))) {
		ast_sem_destroy(&manager_sem);
		ast_log(LOG_ERROR, "Could not create thread: %s\n", strerror(errcode));
		return 0;
	} else {
		return 1;
	}
}

static void stop_table_manager(void)
{
	ast_sem_post(&manager_sem);
	pthread_join(manager_thread, NULL);
	manager_thread = AST_PTHREADT_NULL;
	ast_sem_destroy(&manager_sem);
}

/* Converting from const pointer to non-const (SQLPOINTER) is SAFE.
 *
 * SQLBindParameter is designed for bi-directional data transfer.
 * Logically there is no reason why a driver would clobber memory
 * without asking and I have confirmed this hypothesis with the help
 * of valgrind and a test. Valgrind did not report any memory
 * violations and no segfault occured trying to write to
 * write-protected memory.
 *
 * I also found that leaving all length-related parameters is fine.
 * The driver is able to deduce the size of our parameters just fine
 * from the value and parameter types alone. */

static void bind_parameter(
	SQLHSTMT stmt,
	const SQL_TIMESTAMP_STRUCT *event_time,
	const struct ast_cel_event_record *record,
	const struct tables *table,
	int index,
	int rva
) {
	if (rva == offsetof(struct ast_cel_event_record, event_time)) {
		/* Special case for timestamp. */
		SQLBindParameter(stmt, index, SQL_PARAM_INPUT, SQL_C_TIMESTAMP,
			SQL_TYPE_TIMESTAMP, 0, 0, (SQLPOINTER) event_time, 0, NULL);
	} else if (rva == offsetof(struct ast_cel_event_record, event_type)) {
		/* Special case for event type. */
		if (!cel_show_user_def && record->event_type == AST_CEL_USER_DEFINED) {
			SQLBindParameter(stmt, index, SQL_PARAM_INPUT, SQL_C_CHAR,
				SQL_CHAR, 0, 0, (SQLPOINTER) record->user_defined_name, 0, NULL);
		} else {
			SQLBindParameter(stmt, index, SQL_PARAM_INPUT, SQL_C_CHAR,
				SQL_CHAR, 0, 0, (SQLPOINTER) record->event_name, 0, NULL);
		}
	} else if (rva == offsetof(struct ast_cel_event_record, amaflag)) {
		/* Special case for flags. */
		SQLBindParameter(stmt, index, SQL_PARAM_INPUT, SQL_C_ULONG,
			SQL_INTEGER, 0, 0, (SQLPOINTER) &record->amaflag, 0, NULL);
	} else {
		/* Obtain pointer to string value by adding rva offset to record base address */
		char ** string_ptr = (char **) (((char *) record) + rva);
		SQLBindParameter(stmt, index, SQL_PARAM_INPUT, SQL_C_CHAR,
			SQL_CHAR, 0, 0, (SQLPOINTER) *string_ptr, 0, NULL);
	}
}

static void bind_parameters(
	SQLHSTMT stmt,
	const SQL_TIMESTAMP_STRUCT *event_time,
	const struct ast_cel_event_record *record,
	const struct tables *table
) {
	struct columns *column;
	int index = 1;

	AST_LIST_TRAVERSE(&(table->columns), column, list) {
		if (column->staticval) {
			SQLBindParameter(stmt, index++, SQL_PARAM_INPUT, SQL_C_CHAR,
				SQL_CHAR, 0, 0, (SQLPOINTER) column->staticval, 0, NULL);
		} else if  (column->rva >= 0) {
			bind_parameter(stmt, event_time, record, table, index++, column->rva);
		}
	}
}

static SQLHSTMT prepare_query(
	struct odbc_obj *obj,
	void *data
) {
	int res, i;
	const struct prepare_params *params = data;
	SQLHSTMT stmt;
	SQLINTEGER nativeerror = 0, numfields = 0;
	SQLSMALLINT diagbytes = 0;
	unsigned char state[10], diagnostic[512];

	res = SQLAllocHandle (SQL_HANDLE_STMT, obj->con, &stmt);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		ast_log(LOG_WARNING, "SQL Alloc Handle failed!\n");
		return NULL;
	}

	ast_debug(3, "Executing SQL statement: [%s]\n", params->table->query);
	bind_parameters(stmt, &params->event_time, &params->record, params->table);
	res = ast_odbc_prepare(obj, stmt, params->table->query);

	if (res == SQL_SUCCESS || res == SQL_SUCCESS_WITH_INFO) {
		return stmt;
	} else {
		ast_log(LOG_WARNING, "SQL Prepare failed![%s]\n", params->table->query);
		SQLGetDiagField(SQL_HANDLE_STMT, stmt, 1, SQL_DIAG_NUMBER, &numfields, SQL_IS_INTEGER, &diagbytes);
		for (i = 0; i < numfields; i++) {
			SQLGetDiagRec(SQL_HANDLE_STMT, stmt, i + 1, state, &nativeerror, diagnostic, sizeof(diagnostic), &diagbytes);
			ast_log(LOG_WARNING, "SQL Execute returned an error %d: %s: %s (%d)\n", res, state, diagnostic, diagbytes);
			if (i > 10) {
				ast_log(LOG_WARNING, "Oh, that was good.  There are really %d diagnostics?\n", (int)numfields);
				break;
			}
		}
		SQLFreeHandle (SQL_HANDLE_STMT, stmt);
		return NULL;
	}
}

static int event_filtered_out(
	const struct tables *table,
	const struct ast_cel_event_record *record
) {
	/* Adding a filter means that only events matching the filter will
	 * be written to the table. Multiple filters behave like a logical AND,
	 * meaning the event must match all filters. */

	struct filters *filter;

	AST_LIST_TRAVERSE(&(table->filters), filter, list) {
		if (filter->rva == offsetof(struct ast_cel_event_record, event_time)) {
			/* Filtering timestamps is ignored because pointless. */
		} else if (filter->rva == offsetof(struct ast_cel_event_record, event_type)) {
			/* Special case for event type. */
			if (!cel_show_user_def && record->event_type == AST_CEL_USER_DEFINED) {
				if (!ast_strings_equal(record->user_defined_name, filter->value)) {
					return 1;
				}
			} else {
				if (!ast_strings_equal(record->event_name, filter->value)) {
					return 1;
				}
			}
		} else if (filter->rva == offsetof(struct ast_cel_event_record, amaflag)) {
			/* Special case for flags. */
			int len = snprintf(NULL, 0, "%u", record->amaflag) + 1;
			char * buffer = ast_alloca(len);
			snprintf(buffer, len, "%u", record->amaflag);
			if (!ast_strings_equal(buffer, filter->value)) {
				return 1;
			}
		} else if (filter->rva >= 0) {
			/* Obtain pointer to CEL string variable by adding rva to
			 * base address of CEL record. Then obtain string value by
			 * dereferencing that pointer. */
			const char * value = *(const char **) (((const char *) record) + filter->rva);
			if (!ast_strings_equal(value, filter->value)) {
				return 1;
			}
		}
	}

	return 0;
}

static void timeval_to_sql_timestamp(
	SQL_TIMESTAMP_STRUCT * timestamp,
	const struct timeval * event_time,
	const int usegmtime
) {
	struct timeval date_tv = *event_time;
	struct ast_tm tm = { 0, };
	ast_localtime(&date_tv, &tm, usegmtime ? "UTC" : NULL);

	timestamp->year = tm.tm_year + 1900;
	timestamp->month = tm.tm_mon + 1;
	timestamp->day = tm.tm_mday;
	timestamp->hour = tm.tm_hour;
	timestamp->minute = tm.tm_min;
	timestamp->second = tm.tm_sec;
	timestamp->fraction = tm.tm_usec;
}

static void odbc_log(struct ast_event *event)
{
	struct tables *table;
	struct odbc_obj *obj;
	SQLHSTMT stmt = NULL;
	SQLLEN rows = 0;
	struct prepare_params params;

	params.record.version = AST_CEL_EVENT_RECORD_VERSION;

	if (ast_cel_fill_record(event, &params.record)) {
		return;
	}

	if (AST_RWLIST_RDLOCK(&odbc_tables)) {
		ast_log(LOG_ERROR, "Unable to lock table list.  Insert CEL(s) failed.\n");
		return;
	}

	AST_LIST_TRAVERSE(&odbc_tables, table, list) {

		if (ast_strlen_zero(table->query)) {
			continue;
		}

		if (event_filtered_out(table, &params.record)) {
			continue;
		}

		/* No need to check the connection now; we'll handle any failure in prepare_and_execute */
		if (!(obj = ast_odbc_request_obj(table->connection, 0))) {
			ast_log(LOG_WARNING, "Unable to retrieve database handle for '%s:%s'.  CEL failed: %s\n", table->connection, table->name, table->query);
			continue;
		}

		params.table = table;
		timeval_to_sql_timestamp(&params.event_time, &params.record.event_time, table->usegmtime);

		stmt = ast_odbc_prepare_and_execute(obj, prepare_query, &params);
		if (stmt) {
			SQLRowCount(stmt, &rows);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		}
		if (rows == 0) {
			ast_log(LOG_WARNING, "Insert failed on '%s:%s'.  CEL failed: %s\n", table->connection, table->name, table->query);
		}
		ast_odbc_release_obj(obj);
	}
	AST_RWLIST_UNLOCK(&odbc_tables);
}

static int unload_module(void)
{
	ast_cel_backend_unregister(ODBC_BACKEND_NAME);
	stop_table_manager();
	destroy_tables_safely();
	AST_RWLIST_HEAD_DESTROY(&odbc_tables);

	return 0;
}

static int load_module(void)
{
	AST_RWLIST_HEAD_INIT(&odbc_tables);

	if (ast_cel_backend_register(ODBC_BACKEND_NAME, odbc_log)) {
		ast_log(LOG_ERROR, "Unable to subscribe to CEL events.\n");
		return AST_MODULE_LOAD_DECLINE;
	}

	if (load_config() >= 0 && start_table_manager()) {
		return AST_MODULE_LOAD_SUCCESS;
	} else {
		ast_cel_backend_unregister(ODBC_BACKEND_NAME);
		AST_RWLIST_HEAD_DESTROY(&odbc_tables);
		return AST_MODULE_LOAD_DECLINE;
	}
}

static int reload(void)
{
	destroy_tables_safely();
	load_config();

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, ODBC_BACKEND_NAME,
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload,
	.load_pri = AST_MODPRI_CDR_DRIVER,
);
